@@ -199,7 +199,8 @@ PJ_DECL(pj_str_t) pj_strerror( pj_status_t statcode,
  * @see PJ_PERROR()
  */
 PJ_DECL(void) pj_perror(int log_level, const char *sender, pj_status_t status,
-                        const char *title_fmt, ...);
+                        const char *title_fmt, ...)
+                        PJ_PRINT_FUNC_DECOR(4);
 
 
 /**
@@ -497,7 +498,8 @@ void pj_errno_clear_handlers(void);
     #define pj_perror_wrapper_1(arg)    pj_perror_1 arg
     /** Internal function. */
     PJ_DECL(void) pj_perror_1(const char *sender, pj_status_t status, 
-                              const char *title_fmt, ...);
+                              const char *title_fmt, ...)
+                              PJ_PRINT_FUNC_DECOR(3);
 #else
     #define pj_perror_wrapper_1(arg)
 #endif
@@ -512,7 +514,8 @@ void pj_errno_clear_handlers(void);
     #define pj_perror_wrapper_2(arg)    pj_perror_2 arg
     /** Internal function. */
     PJ_DECL(void) pj_perror_2(const char *sender, pj_status_t status, 
-                              const char *title_fmt, ...);
+                              const char *title_fmt, ...)
+                              PJ_PRINT_FUNC_DECOR(3);
 #else
     #define pj_perror_wrapper_2(arg)
 #endif
@@ -527,7 +530,8 @@ void pj_errno_clear_handlers(void);
     #define pj_perror_wrapper_3(arg)    pj_perror_3 arg
     /** Internal function. */
     PJ_DECL(void) pj_perror_3(const char *sender, pj_status_t status, 
-                              const char *title_fmt, ...);
+                              const char *title_fmt, ...)
+                              PJ_PRINT_FUNC_DECOR(3);
 #else
     #define pj_perror_wrapper_3(arg)
 #endif
@@ -542,7 +546,8 @@ void pj_errno_clear_handlers(void);
     #define pj_perror_wrapper_4(arg)    pj_perror_4 arg
     /** Internal function. */
     PJ_DECL(void) pj_perror_4(const char *sender, pj_status_t status, 
-                              const char *title_fmt, ...);
+                              const char *title_fmt, ...)
+                              PJ_PRINT_FUNC_DECOR(3);
 #else
     #define pj_perror_wrapper_4(arg)
 #endif
@@ -557,7 +562,8 @@ void pj_errno_clear_handlers(void);
     #define pj_perror_wrapper_5(arg)    pj_perror_5 arg
     /** Internal function. */
     PJ_DECL(void) pj_perror_5(const char *sender, pj_status_t status, 
-                              const char *title_fmt, ...);
+                              const char *title_fmt, ...)
+                              PJ_PRINT_FUNC_DECOR(3);
 #else
     #define pj_perror_wrapper_5(arg)
 #endif
@@ -572,7 +578,8 @@ void pj_errno_clear_handlers(void);
     #define pj_perror_wrapper_6(arg)    pj_perror_6 arg
     /** Internal function. */
     PJ_DECL(void) pj_perror_6(const char *sender, pj_status_t status, 
-                              const char *title_fmt, ...);
+                              const char *title_fmt, ...)
+                              PJ_PRINT_FUNC_DECOR(3);
 #else
     #define pj_perror_wrapper_6(arg)
 #endif
