@@ -20,6 +20,7 @@
 #include <pj/errno.h>
 #include <pj/log.h>
 #include <pj/math.h>
+#include <pj/os.h>
 #include <pj/pool.h>
 #include <pj/string.h>
 
@@ -244,6 +245,14 @@ static void ssl_close_sockets(pj_ssl_sock_t *ssock)
 static pj_bool_t on_handshake_complete(pj_ssl_sock_t *ssock, 
                                        pj_status_t status)
 {
+    pj_lock_acquire(ssock->write_mutex);
+    if (ssock->handshake_status != PJ_EUNKNOWN) {
+        pj_lock_release(ssock->write_mutex);
+        return (ssock->handshake_status == PJ_SUCCESS)? PJ_TRUE: PJ_FALSE;
+    }
+    ssock->handshake_status = status;
+    pj_lock_release(ssock->write_mutex);
+
     /* Cancel handshake timer */
     if (ssock->timer.id == TIMER_HANDSHAKE_TIMEOUT) {
         pj_timer_heap_cancel(ssock->param.timer_heap, &ssock->timer);
@@ -261,14 +270,21 @@ static pj_bool_t on_handshake_complete(pj_ssl_sock_t *ssock,
         if (status != PJ_SUCCESS) {
             /* Handshake failed in accepting, destroy our self silently. */
 
-            char buf[PJ_INET6_ADDRSTRLEN+10];
+            char buf1[PJ_INET6_ADDRSTRLEN+10];
+            char buf2[PJ_INET6_ADDRSTRLEN+10];
 
-            if (pj_sockaddr_has_addr(&ssock->rem_addr)) {
-                PJ_PERROR(3,(ssock->pool->obj_name, status,
-                          "Handshake failed in accepting %s",
-                          pj_sockaddr_print(&ssock->rem_addr, buf,
-                                            sizeof(buf), 3)));
-            }
+            if (pj_sockaddr_has_addr(&ssock->local_addr))
+                pj_sockaddr_print(&ssock->local_addr, buf1, sizeof(buf1), 3);
+            else
+                pj_ansi_snprintf(buf1, sizeof(buf1), "(unknown)");
+
+            if (pj_sockaddr_has_addr(&ssock->rem_addr))
+                pj_sockaddr_print(&ssock->rem_addr, buf2, sizeof(buf2), 3);
+            else
+                pj_ansi_snprintf(buf2, sizeof(buf2), "(unknown)");
+
+            PJ_PERROR(3,(ssock->pool->obj_name, status,
+                      "Handshake failed on %s in accepting %s", buf1, buf2));
 
             if (ssock->param.cb.on_accept_complete2) {
                 (*ssock->param.cb.on_accept_complete2) 
@@ -357,6 +373,23 @@ static pj_bool_t on_handshake_complete(pj_ssl_sock_t *ssock,
          * reconnect in the callback.
          */
         if (status != PJ_SUCCESS) {
+            char buf1[PJ_INET6_ADDRSTRLEN+10];
+            char buf2[PJ_INET6_ADDRSTRLEN+10];
+
+            if (pj_sockaddr_has_addr(&ssock->local_addr))
+                pj_sockaddr_print(&ssock->local_addr, buf1, sizeof(buf1), 3);
+            else
+                pj_ansi_snprintf(buf1, sizeof(buf1), "(unknown)");
+
+            if (pj_sockaddr_has_addr(&ssock->rem_addr))
+                pj_sockaddr_print(&ssock->rem_addr, buf2, sizeof(buf2), 3);
+            else
+                pj_ansi_snprintf(buf2, sizeof(buf2), "(unknown)");
+
+            PJ_PERROR(3,(ssock->pool->obj_name, status,
+                      "Handshake failed on %s in connecting to %s",
+                      buf1, buf2));
+
             /* Server disconnected us, possibly due to SSL nego failure */
             ssl_reset_sock_state(ssock);
         }
@@ -640,7 +673,7 @@ static void on_timer(pj_timer_heap_t *th, struct pj_timer_entry *te)
 
     switch (timer_id) {
     case TIMER_HANDSHAKE_TIMEOUT:
-        PJ_LOG(1,(ssock->pool->obj_name, "SSL timeout after %d.%ds",
+        PJ_LOG(1,(ssock->pool->obj_name, "SSL timeout after %ld.%lds",
                   ssock->param.timeout.sec, ssock->param.timeout.msec));
 
         on_handshake_complete(ssock, PJ_ETIMEDOUT);
@@ -736,7 +769,7 @@ static pj_bool_t ssock_on_data_read (pj_ssl_sock_t *ssock,
     }
 
     /* See if there is any decrypted data for the application */
-    if (ssock->read_started) {
+    if (data && ssock->read_started) {
         do {
             read_data_t *buf = *(OFFSET_OF_READ_DATA_PTR(ssock, data));
             void *data_ = (pj_int8_t*)buf->data + buf->len;
@@ -745,17 +778,12 @@ static pj_bool_t ssock_on_data_read (pj_ssl_sock_t *ssock,
 
             status_ = ssl_read(ssock, data_, &size_);
 
-            if (size_ > 0 || status != PJ_SUCCESS) {
+            if (size_ > 0) {
                 if (ssock->param.cb.on_data_read) {
                     pj_bool_t ret;
                     pj_size_t remainder_ = 0;
 
-                    if (size_ > 0)
-                        buf->len += size_;
-                
-                    if (status != PJ_SUCCESS) {
-                        ssock->ssl_state = SSL_STATE_ERROR;
-                    }
+                    buf->len += size_;
 
                     ret = (*ssock->param.cb.on_data_read)(ssock, buf->data,
                                                           buf->len, status,
@@ -775,10 +803,13 @@ static pj_bool_t ssock_on_data_read (pj_ssl_sock_t *ssock,
                  * been signalled to the application along with any remaining
                  * buffer. So, let's just reset SSL socket now.
                  */
+                /*
+                // This has been handled in on_error
                 if (status != PJ_SUCCESS) {
                     ssl_reset_sock_state(ssock);
                     return PJ_FALSE;
                 }
+                */
 
             } else if (status_ == PJ_SUCCESS) {
                 break;
@@ -849,8 +880,8 @@ static pj_bool_t ssock_on_data_sent (pj_ssl_sock_t *ssock,
     pj_ioqueue_op_key_t *app_key = wdata->app_key;
     pj_ssize_t sent_len;
 
-    sent_len = (sent > 0)? wdata->plain_data_len : sent;
-    
+    sent_len = (sent > 0)? (pj_ssize_t)wdata->plain_data_len : sent;
+
     /* Update write buffer state */
     pj_lock_acquire(ssock->write_mutex);
     free_send_data(ssock, wdata);
@@ -919,7 +950,7 @@ static pj_bool_t ssock_on_accept_complete (pj_ssl_sock_t *ssock_parent,
                                            int src_addr_len,
                                            pj_status_t accept_status)
 {
-    pj_ssl_sock_t *ssock;
+    pj_ssl_sock_t *ssock = NULL;
 #ifndef SSL_SOCK_IMP_USE_OWN_NETWORK
     pj_activesock_cb asock_cb;
     pj_activesock_cfg asock_cfg;
@@ -931,6 +962,10 @@ static pj_bool_t ssock_on_accept_complete (pj_ssl_sock_t *ssock_parent,
     PJ_UNUSED_ARG(newconn);
 #endif
 
+    /* Set close-on-exec flag */
+    if (ssock_parent->newsock_param.sock_cloexec)
+        pj_set_cloexec_flag((int)newsock);
+
     if (accept_status != PJ_SUCCESS) {
         if (ssock_parent->param.cb.on_accept_complete2) {
             (*ssock_parent->param.cb.on_accept_complete2)(ssock_parent, NULL,
@@ -1099,7 +1134,6 @@ static pj_bool_t ssock_on_accept_complete (pj_ssl_sock_t *ssock_parent,
                                                    ssock->param.grp_lock);
         if (status != PJ_SUCCESS) {
             ssock->timer.id = TIMER_NONE;
-            status = PJ_SUCCESS;
         }
     }
 
@@ -1398,7 +1432,7 @@ PJ_DEF(pj_status_t) pj_ssl_sock_create (pj_pool_t *pool,
     pj_pool_t *info_pool;
 
     PJ_ASSERT_RETURN(pool && param && p_ssock, PJ_EINVAL);
-    PJ_ASSERT_RETURN(param->sock_type == pj_SOCK_STREAM(), PJ_ENOTSUP);
+    PJ_ASSERT_RETURN((param->sock_type & 0xF) == pj_SOCK_STREAM(), PJ_ENOTSUP);
 
     info_pool = pj_pool_create(pool->factory, "ssl_chain%p", 512, 512, NULL);
     pool = pj_pool_create(pool->factory, "ssl%p", 512, 512, NULL);
@@ -1413,6 +1447,7 @@ PJ_DEF(pj_status_t) pj_ssl_sock_create (pj_pool_t *pool,
     ssock->ssl_state = SSL_STATE_NULL;
     ssock->circ_buf_input.owner = ssock;
     ssock->circ_buf_output.owner = ssock;
+    ssock->handshake_status = PJ_EUNKNOWN;
     pj_list_init(&ssock->write_pending);
     pj_list_init(&ssock->write_pending_empty);
     pj_list_init(&ssock->send_pending);
@@ -1546,17 +1581,26 @@ PJ_DEF(pj_status_t) pj_ssl_sock_get_info (pj_ssl_sock_t *ssock,
     
     if (info->established) {
         info->cipher = ssl_get_cipher(ssock);
-
-        /* Verification status */
-        info->verify_status = ssock->verify_status;
     }
 
+    /* Verification status */
+    info->verify_status = ssock->verify_status;
+
     /* Last known SSL error code */
     info->last_native_err = ssock->last_err;
 
     /* Group lock */
     info->grp_lock = ssock->param.grp_lock;
 
+    /* Native SSL object */
+#if defined(PJ_HAS_SSL_SOCK) && PJ_HAS_SSL_SOCK != 0 && \
+    (PJ_SSL_SOCK_IMP == PJ_SSL_SOCK_IMP_OPENSSL)
+    {
+        ossl_sock_t *ossock = (ossl_sock_t *)ssock;
+        info->native_ssl = ossock->ossl_ssl;
+    }
+#endif
+
     return PJ_SUCCESS;
 }
 
@@ -1923,6 +1967,8 @@ pj_ssl_sock_start_accept2(pj_ssl_sock_t *ssock,
         goto on_error;
 #else
     /* Create socket */
+    if (ssock->param.sock_cloexec)
+        ssock->param.sock_type |= pj_SOCK_CLOEXEC();
     status = pj_sock_socket(ssock->param.sock_af, ssock->param.sock_type, 0, 
                             &ssock->sock);
     if (status != PJ_SUCCESS)
@@ -2054,6 +2100,8 @@ PJ_DEF(pj_status_t) pj_ssl_sock_start_connect2(
                      PJ_EINVAL);
 
     /* Create socket */
+    if (ssock->param.sock_cloexec)
+        ssock->param.sock_type |= pj_SOCK_CLOEXEC();
     status = pj_sock_socket(ssock->param.sock_af, ssock->param.sock_type, 0, 
                             &ssock->sock);
     if (status != PJ_SUCCESS)
